/*
 * Copyright (C) 2013 John Graham <johngavingraham@googlemail.com>.
 *
 * This file is part of Pthread Pool.
 *
 * Pthread Pool is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation, either version 2.1 of the
 * License, or (at your option) any later version.
 *
 * Pthread Pool is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with Pthread Pool.  If not, see
 * <http://www.gnu.org/licenses/>.
 */

/*
 * Test the attributes available to a pthread_pool_t.
 */

#include <pthread_pool.h>

#include <string.h>
#include <errno.h>

#include "test.h"

#define NUM_WORKERS 5

/* Recycle the same variables to test the _destroy functions don't
 * leave anything behind. */
pthread_pool_t        pool;
pthread_poolattr_t    attr;
pthread_pool_worker_t workers[NUM_WORKERS];
pthread_pool_task_t   task;

void *task_routine(void *arg)
{
  return arg;
}

/* Test that a pool can be created without attributes. */
void test_noattr(void)
{
  assert(0 == pthread_pool_create(&pool, NULL));
  
  assert(0 == pthread_pool_destroy(&pool));
}

/* Test that a pool will respect the maxworkers attribute. */
void test_maxworkers(void)
{
  int i;
  
  /* Initialize the attribute and create the pool with it. */
  assert(0 == pthread_poolattr_init(&attr));
  assert(0 == pthread_poolattr_setmaxworkers(&attr, NUM_WORKERS - 1));
  assert(0 == pthread_pool_create(&pool, &attr));
  
  /* Destroy the pool so we know we don't keep a reference to it. */
  assert(0 == pthread_poolattr_destroy(&attr));
  memset(&attr, 0, sizeof(attr));
  
  /* Check we can add all the workers we're allowed. */
  for (i = 0 ; i < NUM_WORKERS - 1 ; i++) {
    assert(0 == pthread_pool_worker_init(&workers[i], &pool, NULL));
  }

  /* We should get an error trying to add any more. */
  assert(pthread_pool_worker_init(&workers[NUM_WORKERS - 1], &pool, NULL));
  
  /* Remove all the workers. */
  for (i = 0 ; i < NUM_WORKERS - 1 ; i++) {
    assert(0 == pthread_pool_worker_destroy(&workers[i], &pool));
  }

  /* Destroy the pool. */
  assert(0 == pthread_pool_destroy(&pool));
}

/* Test that a pool will respect the minworkers attribute. */
void test_minworkers(void)
{
  int i;

  /* Initialize the attribute and create the pool with it. */
  assert(0 == pthread_poolattr_init(&attr));
  assert(0 == pthread_poolattr_setminworkers(&attr, NUM_WORKERS - 1));
  assert(0 == pthread_pool_create(&pool, &attr));
  
  /* Destroy the pool so we know we don't keep a reference to it. */
  assert(0 == pthread_poolattr_destroy(&attr));
  memset(&attr, 0, sizeof(attr));
  
  /* Check we can add all the workers we can. */
  for (i = 0 ; i < NUM_WORKERS ; i++)
    assert(0 == pthread_pool_worker_init(&workers[i], &pool, NULL));
  
  /* We should be able to remove one, but not any others. */
  assert(0 == pthread_pool_worker_destroy(&workers[NUM_WORKERS - 1], &pool));
  assert(pthread_pool_worker_destroy(&workers[NUM_WORKERS - 2], &pool));
  
  /* Destroy the pool (this should remove all workers for us). */
  assert(0 == pthread_pool_destroy(&pool));
}

/* Test that a pool will respect the detectstarve attribute. */
void test_detectstarve(void)
{
  /* Initialize the attribute and create the pool with it. */
  assert_equal(0, pthread_poolattr_init(&attr));
  assert_equal(0, pthread_poolattr_setdetectstarve(&attr, 1));
  assert_equal(0, pthread_pool_create(&pool, &attr));
  
  /* Destroy the pool so we know we don't keep a reference to it. */
  assert_equal(0, pthread_poolattr_destroy(&attr));
  memset(&attr, 0, sizeof(attr));
  
  /* Attempting to join should result in ESRCH, since there are
   * no completed tasks to get imediately and no workers. */
  assert_equal(ESRCH, pthread_pool_join(&pool, NULL));
  
  /* An attempt to wait should succeed, since we don't need to wait
   * for anything. */
  assert_equal(0, pthread_pool_wait(&pool));
  
  /* Add a worker. */
  assert_equal(0, pthread_pool_worker_init(&workers[0], &pool, NULL));
  
  /* An attempt to join should still result in ESRCH, since there
   * are no tasks in the queue for the worker. */
  assert_equal(ESRCH, pthread_pool_join(&pool, NULL));
  
  /* Add one task. */
  assert_equal(0, pthread_pool_task_init(&task, &pool, &task_routine, NULL));
  
  /* Join should now work - but only once. */
  assert_equal(0,     pthread_pool_join(&pool, NULL));
  assert_equal(ESRCH, pthread_pool_join(&pool, NULL));
}

int main(void)
{
  test_noattr();
  test_maxworkers();
  test_minworkers();
  test_detectstarve();
  
  return 0;
}






























