/*
 * Copyright (C) 2013 John Graham <johngavingraham@googlemail.com>.
 *
 * This file is part of Pthread Pool.
 *
 * Pthread Pool is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation, either version 2.1 of the
 * License, or (at your option) any later version.
 *
 * Pthread Pool is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with Pthread Pool.  If not, see
 * <http://www.gnu.org/licenses/>.
 */

#include <pthread_pool.h>

#include <errno.h>
#include <stdio.h>
#include <stdlib.h>

#include <sys/time.h>
#include <unistd.h>

#include "test.h"



/* Maximum number of threads we have creating/destroying tasks. */
#define NUM_THREADS 500



/* The task to perform in a worker thread - just sleep for a bit. */
static void *test_task(void *arg)
{
  usleep(random() % (1 + (intptr_t)arg));
  return arg;
}



/* Thread that randomly adds/removes tasks to/from the pool given as
 * its argument. */
static void *task_minipulator(void *pool_)
{
  pthread_pool_t *pool = pool_;
  int i, n;

  while (random() % 10)
  {
    n = random() % 10;
    for (i = 0 ; i < n ; i++)
    {
      assert(0 ==
        pthread_pool_task_init(
          malloc(sizeof(pthread_pool_task_t)),
                              pool, &test_task,
                              (void *)(random() % 10)));
    }

    /* Join fewer tasks (don't divide zero). */
    if (n)
      n = n / 2;
    for (i = 0 ; i < n ; i++)
    {
      pthread_pool_task_t *task;
      void *arg, *retval;

      int joinret;
      while (ESRCH == (joinret = pthread_pool_join(pool, &task)))
        /* No-op. */;
      assert(0 == joinret);

      assert(0 == pthread_pool_task_getarg(task, &arg));
      assert(0 == pthread_pool_task_getretval(task, &retval));

      assert(arg == retval);

      free(task);
    }
  }

  return NULL;
}



int main()
{
  int i;
  pthread_t threads[NUM_THREADS];
  pthread_pool_worker_t workers[NUM_THREADS];
  pthread_pool_t pool;

  {
    struct timeval tv;
    (void)gettimeofday(&tv, NULL);
    printf("  Random seed: %lu\n", tv.tv_usec);
    srandom(tv.tv_usec);
  }

  assert(0 == pthread_pool_create(&pool, NULL));

  for (i = 0 ; i < NUM_THREADS ; i++)
  {
    assert(0 == pthread_create(&threads[i], NULL,
                               &task_minipulator, &pool));
  }

  /* For each thread add a worker and then join it. */
  for (i = 0 ; i < NUM_THREADS ; i++)
  {
    assert(0 ==
      pthread_pool_worker_init(
        &workers[i], &pool, NULL));

    assert(0 == pthread_join(threads[i], NULL));
  }

  /* All threads have now completed - wait for remaining work
   * to finish. */
  pthread_pool_destroy(&pool);

  return 0;
}






























