/*
 * Copyright (C) 2013 John Graham <johngavingraham@googlemail.com>.
 *
 * This file is part of Pthread Pool.
 *
 * Pthread Pool is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation, either version 2.1 of the
 * License, or (at your option) any later version.
 *
 * Pthread Pool is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with Pthread Pool.  If not, see
 * <http://www.gnu.org/licenses/>.
 */

#include <pthread_pool.h>

#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <unistd.h>

#include "test.h"



#define NUM_DRIVERS        10
#define WORKERS_PER_DRIVER 10



/*
 * Yes, I know this is bad practice etc. and we should issue a full
 * memory barrier, but this works for a boolean flag in practice, at
 * least for now.
 */
volatile int drivers_run = 1;

void *double_task(void *arg)
{
  return (void *)(2 * (intptr_t)arg);
}

void *(*random_task())(void *)
{
  switch (random() % 1)
  {
  case 0:
    return &double_task;
  }

  return &double_task;
}



/**
 * Randomly creates workers and tasks.
 */
void *driver_thread(void *pool_)
{
  int i;
  pthread_pool_t *pool = pool_;
  char buf[200] __attribute((unused));
  pthread_pool_worker_t workers[WORKERS_PER_DRIVER + 20];

  /* Index of next worker to add.
   * WORKERS_PER_DRIVER => workers[] full,
   * 0 => workers[] empty. */
  int workeri = WORKERS_PER_DRIVER / 2; /* Start with 1/2 active. */

  for (i = 0 ; i < workeri ; i++)
    assert(0 == pthread_pool_worker_init(&workers[i], pool, NULL));

  while (drivers_run)
  {
    /* Randomly add or remove a worker. */
    if (random() % 2)
    {
      /* Add worker (if not at max). */
      if (workeri < WORKERS_PER_DRIVER)
      {
        int ret;
        assert(workeri >= 0 && workeri < WORKERS_PER_DRIVER);
        ret = pthread_pool_worker_init(&workers[workeri],
                                       pool, NULL);

        switch (ret)
        {
        case 0:
          workeri++;
          break;

        default:
          /*printf("  (Could not add worker: %s)\n",
                 strerror(ret));*/
          assert(0);
        }
      } /*else printf("No add - workers full\n"); */
    }
    else
    {
      /* Remove worker (if any to remove) - allow ERANGE. */
      if (workeri > 0)
      {
        int ret;
        assert(workeri-1 >= 0 && workeri-1 < WORKERS_PER_DRIVER);
        ret = pthread_pool_worker_destroy(&workers[workeri-1], pool);

        switch (ret)
        {
        case 0:
          workeri--;
          break;

        case ERANGE:
          assert(workeri == 1);
          /*printf("  (Could not remove worker: ERANGE)\n");*/
          break;

        default:
          /*printf("  (Could not remove worker: %s)\n",
                 strerror(ret));*/
          assert(0);
        }
      } /*else printf("No remove - no workers\n");*/
    }
  }

  while (workeri)
  {
    int ret = pthread_pool_worker_destroy(&workers[workeri-1], pool);

    switch (ret)
    {
    case 0:
      workeri--;
      break;

    case ERANGE:
      /*printf("  (ERANGE: Destroying pool...)\n");*/
      assert(0 == pthread_pool_destroy(pool));
      return NULL;

    default:
        /*printf("  (Could not remove worker: %s)\n",
               strerror(ret));*/
        assert(0);
    }
  }

  /*printf("Driver thread done (workeri = %d)\n", workeri);*/
  return NULL;
}



int main(void)
{
  int i;
  pthread_pool_t pool;
  pthread_poolattr_t poolattr;
  pthread_t drivers[NUM_DRIVERS];

  /* With these attributes we *should* always have at least 1 worker,
   * and so never have starvation. */
  assert(0 == pthread_poolattr_init(&poolattr));
  assert(0 == pthread_poolattr_setdetectstarve(&poolattr, 1));
  assert(0 == pthread_poolattr_setminworkers(&poolattr, 1));

  assert(0 == pthread_pool_create(&pool, &poolattr));

  for (i = 0 ; i < NUM_DRIVERS ; i++)
    assert(0 == pthread_create(&drivers[i], NULL,
                               &driver_thread, &pool));

  sleep(3);

  drivers_run = 0;

  sleep(1);

  for (i = 0 ; i < NUM_DRIVERS ; i++)
    assert(0 == pthread_join(drivers[i], NULL));

  return 0;
}


































